
#include "iuptreeutil.h"
#include <cstdio>
#include <cstring>
#include <cassert>
#include <algorithm>

//#define DEBUG_TREE_UTIL


/*****************************************************************************\
* Utilities
\*****************************************************************************/


static void copyString(char** dest, const char* src)
{
  if(*dest)
    delete [] *dest;

  if (src)
  {
    *dest = new char[strlen(src) + 1];
    strcpy(*dest, src);
  }
  else
    *dest = NULL;
}


/*****************************************************************************\
* IupLeafWrapper
\*****************************************************************************/


IupNodeWrapper::IupNodeWrapper(const char* name, bool is_leaf, void* data)
{
  _name = 0;
  _parent = 0;
  _data = data;
  _is_leaf = is_leaf;
  copyString(&_name, name);
}

IupNodeWrapper::~IupNodeWrapper()
{
  if (_name)
  {
//    delete _name;  // Not done before, application can change this if wanted
    _name = NULL;
  }
}

void IupNodeWrapper::addNode(Ihandle* tree, IupNodeWrapper* parent, int id)
{
  _tree = tree;
  _parent = parent;
  IupTreeSetUserId(_tree, id, this);
}

void IupNodeWrapper::removeNode()
{
	int node_id = id();
	IupTreeSetUserId(_tree, node_id, 0);
  IupTreeSetAttribute(_tree, "DELNODE", node_id, "SELECTED");
	_parent = 0;
}

int IupNodeWrapper::id()
{
  return IupTreeGetId(_tree, this);
}

void IupNodeWrapper::updateName(const char* name)
{
  copyString(&_name, name);
}

void IupNodeWrapper::setName(const char* name)
{
  updateName(name);
  SetAttribute("NAME", _name);
}

void IupNodeWrapper::setImage(const char* name)
{
  SetAttribute("IMAGE", name);
}

void IupNodeWrapper::setMarked(bool mark)
{
  SetAttribute("MARKED", mark? "YES": "NO");
}

bool IupNodeWrapper::isMarked()
{
  return IupTreeGetInt(_tree, "MARKED", id())? true: false;
}

void IupNodeWrapper::setColor(int r,int g,int b)
{
 char color[20];
 sprintf(color,"%d %d %d", r, g, b);
 SetAttribute("COLOR", color);
}

int IupNodeWrapper::getDepth()
{
  return IupTreeGetInt(_tree, "DEPTH", id());
}

void IupNodeWrapper::SetAttribute(const char* name, const char* value)
{
  IupTreeSetAttribute(_tree, name, id(), value);
}

void IupNodeWrapper::SetAttributeHandle(const char* name, Ihandle* ih)
{
  IupTreeSetAttributeHandle(_tree, name, id(), ih);
}

char* IupNodeWrapper::GetAttribute(const char* name)
{
  return IupTreeGetAttribute(_tree, name, id());
}

int IupNodeWrapper::selection(int status)
{
  (void)status;
#ifdef DEBUG_TREE_UTIL
  printf("IupTreeWrapper::selection(status=%d) [%d, %s]\n", status, id(), getName());
#endif
  return IUP_DEFAULT;
}

int IupNodeWrapper::rename(char* name)
{
  (void)name;
#ifdef DEBUG_TREE_UTIL
  printf("IupTreeWrapper::rename(name=%s) [%d, %s]\n", name, id(), getName());
#endif
  return IUP_DEFAULT;
}

int IupNodeWrapper::rightClick()
{
#ifdef DEBUG_TREE_UTIL
  printf("IupTreeWrapper::rightClick() [%d, %s]\n", id(), getName());
#endif
  return IUP_DEFAULT;
}

int IupNodeWrapper::showRename()
{
#ifdef DEBUG_TREE_UTIL
  printf("IupTreeWrapper::showRename() [%d, %s]\n", id(), getName());
#endif
  return IUP_DEFAULT;
}


/*****************************************************************************\
* IupLeafWrapper
\*****************************************************************************/


int IupLeafWrapper::execute(char* name)
{
  (void)name;
#ifdef DEBUG_TREE_UTIL
  printf("IupTreeWrapper::execute(name=%s) [%d, %s]\n", name, id(), getName());
#endif
  return IUP_DEFAULT;
}


/*****************************************************************************\
* Callbacks
\*****************************************************************************/


static int IupTreeWrapper_RightClick(Ihandle* ih, int id)
{
  IupNodeWrapper* node = (IupNodeWrapper*)IupTreeGetUserId(ih, id);
  if (node) return node->rightClick();
  return IUP_DEFAULT;
}

static int IupTreeWrapper_Selection(Ihandle* ih, int id, int status)
{
  IupNodeWrapper* node = (IupNodeWrapper*) IupTreeGetUserId(ih, id);
  if (node) return node->selection(status);
  return IUP_DEFAULT;
}

static int IupTreeWrapper_BranchOpen(Ihandle* ih, int id)
{
  IupNodeWrapper* node = (IupNodeWrapper*) IupTreeGetUserId(ih, id);
  if (node && !node->isLeaf())
  {
    IupTreeWrapper* branch = (IupTreeWrapper*)node;
    return branch->branchOpen();
  }
  return IUP_DEFAULT;
}

static int IupTreeWrapper_BranchClose(Ihandle* ih, int id)
{
  IupNodeWrapper* node = (IupNodeWrapper*) IupTreeGetUserId(ih, id);
  if (node && !node->isLeaf())
  {
    IupTreeWrapper* branch = (IupTreeWrapper*)node;
    return branch->branchClose();
  }
  return IUP_DEFAULT;
}

static int IupTreeWrapper_ExecuteLeaf(Ihandle* ih, int id)
{
  IupNodeWrapper* node = (IupNodeWrapper*)IupTreeGetUserId(ih, id);
  if (node && node->isLeaf())
  {
    IupLeafWrapper* leaf = (IupLeafWrapper*)node;
    return leaf->execute(node->getName());
  }
  return IUP_DEFAULT;
}

static int IupTreeWrapper_Rename(Ihandle* ih, int id, char* name)
{
  int ret = IUP_DEFAULT;
  IupNodeWrapper* node = (IupNodeWrapper*) IupTreeGetUserId(ih, id);
  if (node) 
  {
    ret = node->rename(name);
    if (ret == IUP_DEFAULT)
      node->updateName(name);
  }
  return ret;
}

static int IupTreeWrapper_ShowRename(Ihandle* ih, int id)
{
  IupNodeWrapper* node = (IupNodeWrapper*) IupTreeGetUserId(ih, id);
  if (node) return node->showRename();
  return IUP_DEFAULT;
}


/*****************************************************************************\
* IupLeafWrapper
\*****************************************************************************/


IupTreeWrapper::IupTreeWrapper(Ihandle* tree, const char* name, void* data)
  :IupNodeWrapper(name, false, data)
{
  _tree = tree;
  IupTreeSetUserId(_tree, 0, this); 
  IupSetAttribute(_tree, "NAME0", _name);
//  IupSetAttribute(_tree, "MARKMODE", "MULTIPLE");  // Not done before, application can change this if wanted

  IupSetAttribute(_tree, "WRAPPER", (char*) this);

  IupSetCallback(_tree, "RIGHTCLICK_CB", (Icallback) IupTreeWrapper_RightClick);
  IupSetCallback(_tree, "SELECTION_CB", (Icallback) IupTreeWrapper_Selection);
  IupSetCallback(_tree, "BRANCHOPEN_CB", (Icallback) IupTreeWrapper_BranchOpen);
  IupSetCallback(_tree, "BRANCHCLOSE_CB", (Icallback) IupTreeWrapper_BranchClose);
  IupSetCallback(_tree, "EXECUTELEAF_CB", (Icallback) IupTreeWrapper_ExecuteLeaf);
  IupSetCallback(_tree, "RENAME_CB", (Icallback) IupTreeWrapper_Rename);
  IupSetCallback(_tree, "SHOWRENAME_CB", (Icallback) IupTreeWrapper_ShowRename);
}

void IupTreeWrapper::addLeafFront(IupLeafWrapper* leaf)
{
  // Adds a leaf to the start of the list of leaves,
  // this is after the last branch of the list of SubTrees  
  int pos = leafFrontPosition();

  if(_subTrees.empty())
  {
    // Adds a leaf as the first node of the current branch, 
    // pos-1 is the current branch.
    IupTreeSetAttribute(_tree, "ADDLEAF", pos - 1, leaf->getName());
  }
  else
  {
    // Adds a leaf after the last branch of the current branch, 
    // pos-1 is the last branch.
    IupTreeSetAttribute(_tree, "INSERTLEAF", pos - 1, leaf->getName());  
  }

  if(!_subTrees.empty())
  {
    // If used INSERT then must retrieve the actual pos of the new leaf
    pos = IupGetInt(_tree, "LASTADDNODE");
  }

  leaf->addNode(_tree, this, pos);

  _leaves.push_front(leaf);
}

void IupTreeWrapper::addLeafBack(IupLeafWrapper* leaf)
{
  // Adds a leaf to the end of the list of leaves
  int pos = leafBackPosition();

  if(_subTrees.empty() && _leaves.empty())
  {
    // Adds a leaf as the first node of the current branch, 
    // pos-1 is the current branch.
    IupTreeSetAttribute(_tree, "ADDLEAF", pos - 1, leaf->getName());
  }
  else
  {
    // Adds a leaf after the last leaf of the current branch, 
    // pos-1 is the last leaf, or the last branch.
    IupTreeSetAttribute(_tree, "INSERTLEAF", pos - 1, leaf->getName());
  }

  if (!(_subTrees.empty() && _leaves.empty()))
  {
    // If used INSERT then must retrieve the actual pos of the new leaf
    pos = IupGetInt(_tree, "LASTADDNODE");
  }

  leaf->addNode(_tree, this, pos);

  _leaves.push_back(leaf);
}

void IupTreeWrapper::addSubTreeFront(IupTreeWrapper* subTree)
{
  // Adds a subtree to the start of the list of SubTrees
  int pos = subTreeFrontPosition();

  // Adds a branch as the first node of the current branch, 
  // pos-1 is the current branch. 
  IupTreeSetAttribute(_tree, "ADDBRANCH", pos - 1, subTree->getName());   

  subTree->addNode(_tree, this, pos);

  _subTrees.push_front(subTree);
}

void IupTreeWrapper::addSubTreeBack(IupTreeWrapper* subTree)
{
  // Adds a subtree to the end of the list of SubTrees
  int pos = subTreeBackPosition();

  if(_subTrees.empty())
  {
    // Adds a branch as the first node of the current branch, 
    // pos-1 is the current branch.
    IupTreeSetAttribute(_tree, "ADDBRANCH", pos - 1, subTree->getName());
  }
  else
  {
    // Adds a branch after the last branch of the current branch, 
    // pos-1 is the last branch.
    IupTreeSetAttribute(_tree, "INSERTBRANCH", pos - 1, subTree->getName());
  }

  if(!_subTrees.empty())
  {
    // If used INSERT then must retrieve the actual pos of the new branch
    pos = IupGetInt(_tree, "LASTADDNODE");
  }

  subTree->addNode(_tree, this, pos);

  _subTrees.push_back(subTree);
}

//////////////////////////////////////////////////////////
// IupTreeWrapper organizes all branches before all leaves
// IupTreeWrapper indexing starts at zero, but does not includes the root, 
// so must increment 1 to match IupTree indexing.
//////////////////////////////////////////////////////////

int IupTreeWrapper::subTreeFrontPosition()
{
  // the Id of the first node of the current branch, 
  // it is the id of the first subTree in IupTreeWrapper
  return id() + 1;
}

int IupTreeWrapper::subTreeBackPosition()
{
  // if empty add at Front
  if(_subTrees.empty())
    return subTreeFrontPosition();

  // the Id of the last subTree in IupTreeWrapper
  IupTreeWrapper* p = _subTrees.back();
  return p->id() + 1;
}

int IupTreeWrapper::leafFrontPosition()
{
  // the first leaf starts after the last branch in IupTreeWrapper
  return subTreeBackPosition();
}

int IupTreeWrapper::leafBackPosition()
{
  // if empty add at Front
  if(_leaves.empty())
    return leafFrontPosition();

  // the Id of the last leaf in IupTreeWrapper
  IupLeafWrapper* p = _leaves.back();
  return p->id() + 1;
}

IupLeafWrapper* IupTreeWrapper::findLeaf(void* data, bool searchSubTrees)
{
  LeafList::iterator i;
  for(i = _leaves.begin(); i != _leaves.end(); i++)
  {
    if((*i)->getData() == data)
      return *i;
  }

  if(searchSubTrees)
  {
    TreeList::iterator j;
    for(j = _subTrees.begin(); j != _subTrees.end(); j++)
    {
      IupLeafWrapper* node = (*j)->findLeaf(data, searchSubTrees);

      if(node != 0)
        return node;
    }
  }

  return 0;
}

bool IupTreeWrapper::removeLeaf(IupLeafWrapper* leaf)
{
  LeafList::iterator i = std::find(_leaves.begin(), _leaves.end(), leaf);

  if (i == _leaves.end())
    return false;

  leaf->removeNode();

  _leaves.erase(i);

  return true;
}


bool IupTreeWrapper::removeLeaves(bool includeSubTrees, bool deleteLeaves)
{
  if(includeSubTrees)
  {
    TreeList::iterator i;
    for(i = _subTrees.begin(); i != _subTrees.end(); i++)
      (*i)->removeLeaves(includeSubTrees, deleteLeaves);
  }

  LeafList::iterator j;
  for(j = _leaves.begin(); j != _leaves.end(); j++)
  {
    IupLeafWrapper* leaf = *j;

    leaf->removeNode();

    if (deleteLeaves)
      delete leaf;
  }

  _leaves.clear();

  return true;
}


bool IupTreeWrapper::removeSubTrees(bool deleteSubTrees, bool deleteLeaves)
{
  removeLeaves(true, deleteLeaves);

  TreeList::iterator i;
  for(i = _subTrees.begin(); i != _subTrees.end(); i++)
  {
		IupTreeWrapper* ct = *i;
    
    ct->removeSubTrees(deleteSubTrees, deleteLeaves);

    ct->removeNode();

    if (deleteSubTrees)
      delete ct;
  }

  _subTrees.clear();

  return true;  
}

bool IupTreeWrapper::removeSubTree(IupTreeWrapper* tree, bool deleteSubTrees, bool deleteLeaves)
{
	TreeList::iterator i;
	for(i = _subTrees.begin(); i != _subTrees.end(); i++)
	{
		IupTreeWrapper* ct = *i;

		if (ct == tree)
		{
			ct->removeSubTrees(deleteSubTrees, deleteLeaves);

			ct->removeNode();

			if (deleteSubTrees)
				delete ct;

			_subTrees.erase(i);

			return true;
		}
	}

	return false;
}

int IupTreeWrapper::branchOpen()
{
#ifdef DEBUG_TREE_UTIL
  printf("IupTreeWrapper::branchOpen() [%d, %s]\n", id(), getName());
#endif
  return IUP_DEFAULT;
}

int IupTreeWrapper::branchClose()
{
#ifdef DEBUG_TREE_UTIL
  printf("IupTreeWrapper::branchClose() [%d, %s]\n", id(), getName());
#endif
  return IUP_DEFAULT;
}
